# ------------------------------------------------------------------------------
#  es7s/core
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import getpass
import os
import re
import typing as t
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime, timedelta

import click
import pytermor as pt

from .._decorators import (
    cli_pass_context,
    _catch_and_log_and_exit,
    cli_option,
    cli_argument,
    cli_command,
)
from ...shared import (
    get_stdout,
    get_logger,
    stream_subprocess,
    Styles as BaseStyles,
    init_io,
    init_logger,
)
from ...shared.file import IFileIconRenderer, FileIconRendererFactory, IFile
from ...shared.path import LS_PATH

SGR_RESET = pt.SeqIndex.RESET.assemble()


@cli_command(name=__file__, short_help="list directory contents with bells and whistes")
@cli_argument("file", type=str, required=False, nargs=-1)
@cli_option(
    "-g",
    "--groups",
    is_flag=True,
    help="Display group column",
)
@cli_option(
    "-h",
    "--hard-links",
    is_flag=True,
    help="Display amount of hard links to the file",
)
@cli_option(
    "-L",
    "--dereference",
    is_flag=True,
    help="Follow the symlinks and print actual file properties instead of link file properties.",
)
@cli_option(
    "-o",
    "--octal-perms",
    is_flag=True,
    help="Display extra column with permissions in octal form (e.g., 0644, 0755).",
)
@cli_option(
    "-Q",
    "--quote-names",
    is_flag=True,
    help="Enclose filenames in quotes and escape non-printables.",
)
@cli_option(
    "-r",
    "--reverse",
    is_flag=True,
    help="Reverse sorting order.",
)
@cli_option(
    "-R",
    "--recursive",
    is_flag=True,
    help="Descend into subdirectories.",
)
@cli_option(
    "-s",
    "--sort-by-size",
    is_flag=True,
    help="Sort list by file size (directories are unaffected), biggest last.",
)
@cli_option(
    "-t",
    "--sort-by-time",
    is_flag=True,
    help="Sort list by modification time, newest last.",
)
@cli_option(
    "-U",
    "--unicode-icons",
    is_flag=True,
    help="Use shorter icon set (5 items) which should be supported by literally everything down "
    "to a potato (inclusive).",
)
@cli_pass_context
@_catch_and_log_and_exit
class ListDirCommand:
    """
    Wrapper around GNU 'ls' with preset settings for "one-button" usage.
    FILE is a path(s) to directories and files of interest and can be used
    multiple times, default is current/working directory.

    Involves and formats the output of '/usr/bin/ls'. Default format is long,
    entries are sorted by file name, file sizes are in human-readable SI-form.
    No dereferencing is performed.
    """

    # COUNT_CHILDREN_CMD = [
    #     "find",
    #     "%(file)s",
    #     "-maxdepth", "1",
    #     "-type", "d",
    #     "-exec", "sh", "-c", "find {} -maxdepth 1 | wc -l",
    #     ";",
    #     "-print",
    # ]

    SPLIT_REGEX = re.compile(r"(\d+,\s+\d+|\S+)")
    ATTR_AMOUNT = 8
    OUT_COL_SEPARATOR = " "

    def __init__(
        self,
        ctx: click.Context,
        *,
        file: t.Tuple[str],
        dereference: bool,
        groups: bool,
        hard_links: bool,
        octal_perms: bool,
        quote_names: bool,
        recursive: bool,
        reverse: bool,
        sort_by_size: bool,
        sort_by_time: bool,
        unicode_icons: bool,
        **kwargs,
    ):
        ls_reverse_arg = reverse
        if sort_by_size or sort_by_time:
            ls_reverse_arg = not reverse
        color = "always" if get_stdout().sgr_allowed else "never"
        quoting_style = "shell-escape-always" if quote_names else "literal"

        ls_args = [
            LS_PATH,
            "-l",
            "--classify",
            "--almost-all",
            "--si",
            "--group-directories-first",
            "--time-style=+%s",
            "--color=" + color,
            "--quoting-style=" + quoting_style,
            "-t" if sort_by_time else "",
            "-S" if sort_by_size else "",
            "--reverse" if ls_reverse_arg else "",
            "--recursive" if recursive else "",
            "--dereference" if dereference else "",
            "--no-group" if not groups else "",
        ]
        self._custom_columns = dict(
            octperms=octal_perms,
            hlinks=hard_links,
            group=groups,
        )
        self._icon_renderer = FileIconRendererFactory.make(unicode_icons)
        self._run(file, *filter(None, ls_args))

    def _run(self, file: t.Tuple[str], *ls_args: str):
        ls_args += ("--", *file)

        for out, err in stream_subprocess(*ls_args):
            if out and (line := self._process_ls_output(out.rstrip())):
                get_stdout().echo(line)
            if err:
                get_logger().error(err.rstrip())

    def _process_ls_output(self, line: str) -> str:
        if line.startswith("total"):
            return get_stdout().echo_rendered(_highligher.colorize(line))
        try:
            return self._process_regular_line(line)
        except ValueError as _:
            return line

    def _get_split_limit(self) -> int:
        return self.ATTR_AMOUNT - 1 - (1 if not self._custom_columns.get("group") else 0)

    def _process_regular_line(self, line: str) -> str:
        splitted = list(self._split_ls_line(line))

        if not self._custom_columns.get("group"):
            splitted.insert(3, " ")

        if len(splitted) != self.ATTR_AMOUNT:
            return line

        file = File(*splitted)
        columns = file.render(self._custom_columns, self._icon_renderer)
        filtered_columns = self._assemble_line(columns)
        return self.OUT_COL_SEPARATOR.join(filtered_columns)

    def _assemble_line(self, render_parts: OrderedDict) -> t.Iterable[str]:
        for k, v in render_parts.items():
            if not self._custom_columns.get(k, True):
                continue
            yield v

    def _split_ls_line(self, s: str) -> tuple[str, ...]:
        splitted = self.SPLIT_REGEX.split(s, self._get_split_limit())
        pairs = [iter(splitted)] * 2
        for value in zip(*pairs, strict=False):
            yield "".join(value)  # value ex.: (' ', '461')
        yield splitted[-1]


class _Styles(BaseStyles):
    @staticmethod
    def _make_rwx(r: pt.FT, w: pt.FT, x: pt.FT, base: pt.Style = pt.NOOP_STYLE) -> t.Dict:
        return {k: pt.Style(base, fg=v) for k, v in dict(r=r, w=w, x=x).items()}

    INACTIVE = BaseStyles.TEXT_LABEL
    INACTIVE_ATTR = BaseStyles.TEXT_DISABLED
    OCTAL_PERMS = pt.Style(fg=pt.cv.MAGENTA)
    HARD_LINKS_DIR = INACTIVE
    HARD_LINKS_FILE = pt.Style(fg=pt.cv.RED, bold=True)
    HARD_LINKS_FILE_GT1 = pt.Style(fg=pt.cv.YELLOW, bg=pt.cv.DARK_RED_2, bold=True)
    SPECIAL_ATTR_STICKY = pt.Style(fg=pt.Color256.get_by_code(68), bg=pt.cv.BLACK)
    SPECIAL_ATTR_SETID = pt.Style(fg=pt.Color256.get_by_code(74), bg=pt.cv.BLACK, bold=True)
    # INACTIVE_SPECIAL_ATTR = pt.Style(fg=pt.Color256.get_by_code(62))
    EXTENDED_ATTR = pt.Style(fg="true-white")
    OWNER_COLOR_ROOT = pt.cv.HI_RED
    OWNER_COLOR_CURUSER = pt.NOOP_COLOR
    OWNER_COLOR_OTHER = pt.Color256.get_by_code(187)  # not current and not root
    RWX_DIR_OTHERS_WRITABLE = pt.Style(fg=pt.cv.DARK_GOLDENROD, bg=pt.cv.BLACK)
    RWX_MAP_RGB = {
        "user": _make_rwx("hi-red kalm", "yellow kalm", "hi-yellow kalm", pt.Style(bold=True)),
        "group": _make_rwx(0x794544, 0x79513D, 0x8D7951),
        "others": _make_rwx(0x5F4A49, 0x635149, 0x625E56),
        "dir_others_writable": RWX_DIR_OTHERS_WRITABLE,
    }
    RWX_MAP_256 = {
        "user": _make_rwx(
            pt.cv.INDIAN_RED_3, pt.cv.LIGHT_SALMON_2, pt.cv.LIGHT_GOLDENROD_5, pt.Style(bold=True)
        ),
        "group": _make_rwx(
            pt.cv.LIGHT_PINK_1, pt.cv.NAVAJO_WHITE_1, pt.cv.WHEAT_1, pt.Style(dim=True)
        ),
        "others": _make_rwx(pt.cv.GRAY_30, pt.cv.GRAY_35, pt.cv.GRAY_42),
        "dir_others_writable": RWX_DIR_OTHERS_WRITABLE,
    }
    RWX_MAP_16 = {
        "user": _make_rwx(pt.cv.HI_RED, pt.cv.YELLOW, pt.cv.HI_YELLOW, pt.Style(bold=True)),
        "group": _make_rwx(pt.cv.HI_RED, pt.cv.YELLOW, pt.cv.HI_YELLOW, pt.Style(dim=True)),
        "others": _make_rwx(pt.cv.WHITE, pt.cv.GRAY, pt.cv.GRAY),
        "dir_others_writable": pt.cv.YELLOW,
    }
    RWX_MAP = RWX_MAP_256


_highligher = pt.Highlighter(dim_units=False)


@dataclass
class File(IFile):
    NO_SIZE_PERM_REGEX = re.compile(R"^d")
    UNKNOWN_SIZE_PERM_REGEX = re.compile(R"^l")
    INACTIVE_ATTR_REGEX = re.compile(R"([-?]+)")
    FILE_CLASS_REGEX = re.compile(R"([*/=>@|]?)$")  # there will be \e[m or \e]K
    SGR_SEQ_REGEX = re.compile(R"(\x1b)(\[)([0-9;]*)(m)")

    def __post_init__(self):
        self._inactive_attr_replace = get_stdout().render(R"\1", _Styles.INACTIVE)

        self.name_prefix = ""
        self.name += self.name_extra
        self.name_extra = ""

        if cls_match := self.FILE_CLASS_REGEX.search(self.name):
            self.cls_char = cls_match.group(1) or " "
            self.name = self.name.removesuffix(self.cls_char)

        match self.cls_char:
            case "/":
                self.is_dir = True

        match self.perm[0]:
            case "l":
                self.is_link = True
            case "c":
                self.is_char = True
            case "b":
                self.is_block = True
            case "s":
                self.is_socket = True
            case "p":
                self.is_pipe = True

    def render(self, custom_columns: dict, icon_renderer: IFileIconRenderer) -> OrderedDict:
        stdout = get_stdout()
        filefmt: str = ""

        def found(sgrm: re.Match):
            # to correctly process lines like: 4-Oct␣14:00␣⢸θ⡇⢸ǝF66⡇.aptitude⢸θ⡇/↵
            # adding resetter 0m to fmt will make it useless, so omit it
            nonlocal filefmt
            if sgrm.group(3) in ["", "0"] and len(filefmt) > 0:
                return ""
            filefmt += sgrm.group()
            return ""

        if stdout.renderer.is_format_allowed:
            # extract existing SGRs (from ls) and reapply them to bigger set of fields
            self.name = pt.utilstr.SGR_SEQ_REGEX.sub(found, self.name)
            if name_prefix := re.match(r"^\s+", self.name):
                self.name_prefix = name_prefix.group()
                self.name = self.name[len(self.name_prefix) :]
            if self.name.startswith(".") and self.is_dir:
                filefmt += pt.ansi.SeqIndex.DIM.assemble()

        classfmt_resetters = pt.ansi.SeqIndex.BG_COLOR_OFF
        if pt.contains_sgr(filefmt, pt.ansi.SeqIndex.BLACK.params[0]):
            classfmt_resetters += pt.ansi.SeqIndex.COLOR_OFF
        classfmt = filefmt + classfmt_resetters.assemble()

        iconfmt_resetters = pt.ansi.SeqIndex.ITALIC_OFF
        iconfmt = filefmt + iconfmt_resetters.assemble()

        if self.perm.endswith("+"):
            self.is_special = True
            self.perm = self.perm[:-1]

        perm_render, perm_raw = self._render_perm()

        if perm_raw.count("?") > 3:
            self.is_invalid = True
            if perm_raw.count("?") >= 10:  # :nu_ebana:
                self.is_special = True
                perm_raw = re.sub(R"(\?{9})\?", R"\1", perm_raw)  # (」°ロ°)」
                self.hlinks = " " + self.hlinks
            perm_raw = re.sub(R"\S", "?", perm_raw)  # X_X
            perm_render = stdout.render(perm_raw, _Styles.INACTIVE_ATTR)

        if not custom_columns.get("octperms", False):
            perm_render = " " + perm_render

        return OrderedDict(
            octperms=self._render_oct_perm(perm_raw),
            perms=perm_render,
            hlinks=self._render_hlinks(),
            owner=self._render_user(self.owner),
            group=self._render_group(self.group),
            size=self._render_size(),
            date=self._render_datetime(),
            fclass=(classfmt + self._render_class() + SGR_RESET),
            fname=(
                iconfmt
                + icon_renderer.render(self)
                + filefmt
                + self._render_name(filefmt)
                + SGR_RESET
            ),
        )

    def _render_oct_perm(self, perm_raw: str) -> str:
        if self.is_invalid:
            return " ???"

        def get_st(v: int, pos: int) -> pt.Style:
            if pos > 0 or v > 0:
                return _Styles.OCTAL_PERMS
            return _Styles.TEXT_DISABLED

        perms = pt.Text()
        for n in range(3):
            ppart = perm_raw[n * 3 : (n + 1) * 3]
            ppart_int = 0
            for idx, val in enumerate([4, 2, 1]):
                if ppart[idx].islower():
                    ppart_int += val
            perms.append(pt.Fragment(str(ppart_int), get_st(ppart_int, n + 1)))

        specials = perm_raw[2] + perm_raw[5] + perm_raw[8]
        spart_int = 0
        for idx, val in enumerate([4, 2, 1]):
            if specials[idx].lower() in ("s", "t"):
                spart_int += val
        perms.prepend(pt.Fragment(str(spart_int or " "), get_st(spart_int, 0)))

        return get_stdout().render(perms)

    def _render_perm(self) -> tuple[str, str]:
        raw = self.perm[1:]
        if self.is_special:
            raw += "+"
        else:
            raw += " "

        result = pt.Text()
        for idx, c in enumerate(raw):
            result += pt.Fragment(*self._render_perm_chars(idx, c))
        return get_stdout().render(result), result.render(pt.renderer.NoOpRenderer)

    def _render_perm_chars(self, idx: int, c: str) -> tuple[str, pt.FT]:
        if idx < 3:
            rwx_set = _Styles.RWX_MAP.get("user")
        elif idx < 6:
            rwx_set = _Styles.RWX_MAP.get("group")
        else:
            rwx_set = _Styles.RWX_MAP.get("others")

        if idx >= 6 and self.is_dir and c == "w":
            return c, _Styles.RWX_MAP.get("dir_others_writable")

        match c:
            case "+":
                return c, _Styles.EXTENDED_ATTR
            case "-":
                return c, _Styles.INACTIVE_ATTR
            case "s" | "S":
                if self.is_dir:
                    return c, _Styles.TEXT_DEFAULT

                return c, _Styles.SPECIAL_ATTR_SETID
            case "t" | "T":
                if not self.is_dir:
                    return c, _Styles.TEXT_DEFAULT
                return c, _Styles.SPECIAL_ATTR_STICKY
            case "r" | "w" | "x":
                return c, rwx_set.get(c)
            case " ":  # padding
                return c, pt.NOOP_STYLE
            case _:  # unknown
                return c, _Styles.CRITICAL_ACCENT

    def _render_group(self, group: str) -> str:
        default = self._render_user(group)
        if self.is_invalid:
            return default
        return get_stdout().render(default, pt.Style(dim=True))

    def _render_user(self, user: str) -> str:
        if self.is_invalid:
            return get_stdout().render(
                "?".rjust(len(user) - 1),
                _Styles.INACTIVE,
            )

        def _cur_user() -> str:
            try:
                return os.getlogin()
            except OSError:
                return getpass.getuser()

        owner = user.removeprefix(" ")
        if owner.strip() == "root":
            st = _Styles.OWNER_COLOR_ROOT
        elif owner.strip() == _cur_user():
            st = _Styles.OWNER_COLOR_CURUSER
        else:
            st = _Styles.OWNER_COLOR_OTHER

        return get_stdout().render(owner, st)

    def _render_hlinks(self) -> str:
        if self.is_invalid:
            return "?".rjust(len(self.hlinks))

        hlinks_raw = self.hlinks
        if self.is_special:
            hlinks_raw = " " + hlinks_raw

        hlinks_style = _Styles.INACTIVE
        if not self.is_dir:
            hlinks_style = _Styles.HARD_LINKS_FILE
            if int(self.hlinks) > 1:
                hlinks_style = _Styles.HARD_LINKS_FILE_GT1
        return get_stdout().render(f"{hlinks_raw}", hlinks_style)

    def _render_size(self) -> str:
        def get_inactive_label() -> str | None:
            if self.is_invalid:
                return "?"
            if self.UNKNOWN_SIZE_PERM_REGEX.match(self.perm):
                return "?"
            if self.NO_SIZE_PERM_REGEX.match(self.perm):
                return "-"

        if inactive_label := get_inactive_label():
            return get_stdout().render(inactive_label.rjust(len(self.size)), _Styles.INACTIVE)
        return get_stdout().render(_highligher.colorize(self.size))

    def _render_datetime(self) -> str:
        if self.is_invalid:
            return "?".rjust(13)

        filedt = datetime.fromtimestamp(int(self.timestamp))
        diff = datetime.now() - filedt

        datefmt = pt.NOOP_SEQ
        if get_stdout().renderer.is_format_allowed:
            datefmt = self._get_date_format(diff)

        if diff.days >= 180:
            filedt_str = filedt.strftime(" %e %b  %Y")
        else:
            filedt_str = filedt.strftime(" %e %b %R")
        return f"{datefmt}{filedt_str}{pt.ansi.get_closing_seq(datefmt)}"

    def _render_class(self) -> str:
        if self.is_block:
            return "+"
        if self.is_char:
            return "-"
        if self.is_link:
            return "~" if self.is_dir else "@"
        if self.is_socket:
            return "="
        if self.is_pipe:
            return "|"
        return self.cls_char

    def _render_name(self, filefmt: str) -> str:
        if self.is_link:
            self.name = self.name.replace(" -> ", SGR_RESET + " → " + filefmt)
        return self.name_prefix + self.name

    def _auto_apply_inactive_style(self, string: str) -> str:
        return re.sub(self.INACTIVE_ATTR_REGEX, self._inactive_attr_replace, string)

    @staticmethod
    def _get_date_format(diff: timedelta) -> pt.SequenceSGR:
        if diff < timedelta(hours=1):
            code = 231  # true white
        elif diff < timedelta(days=7):
            code = 254  # 89% gray
        elif diff < timedelta(days=30):
            code = 253 - (3 * diff.days // (30 - 7))  # 253-250 (85-74%)
        elif diff < timedelta(days=365 * 12):
            code = 249 - (diff.days // 365)  # 249-237 (70-23%)
        else:
            code = 237  # 23% gray
        return pt.make_color_256(code)


if __name__ == "__main__":
    init_io()
    init_logger()
    try:
        ListDirCommand().run()
    except Exception as e:
        get_logger().exception(e)
        exit(1)
